/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.speedo.api.Debug;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.resource.NotSupportedException;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.ConnectionSpec;
import javax.resource.cci.RecordFactory;
import javax.resource.cci.ResourceAdapterMetaData;
import javax.resource.spi.ConnectionManager;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOException;
import java.util.Properties;
import java.util.Collection;

/**
 * @author	P. Dechamboux
 */
public class JdoConnectionFactory
        implements ConnectionFactory, ResourceAdapterMetaData,
        PersistenceManagerFactory {
    private boolean started = false;
    /**
     * The logger into which traces about JdoConnectionFactory are produced.
     */
    private Logger logger;
    /**
     * The JDO Adapter Version.
     */
    private final static String JDOADAPTERVERSION = "1.00";
    /**
     * The JDO Adapter Vendor Name (i.e. ObjectWeb !!).
     */
    private final static String JDOADAPTERVENDORNAME = "The ObjectWeb Consortium";
    /**
     * The JDO Adapter Name.
     */
    private final static String JDOADAPTERNAME = "OW Java Data Object (JDO) Adapter";
    /**
     * A short description of the JDO Adapter.
     */
    private final static String JDOADAPTERSHORTDESCRIPTION = "It provides a JCA wrapper to any JDO compliant implementation.";
    /**
     * The JDO Spec Version (none yet).
     */
    private final static String JDOSPECVERSION = "1.00";

    /**
     * The JNDI Reference associated to this adapter.
     */
    private Reference jndiReference = null;
    /**
     * The ConnectionManager used to actually get Connection.
     */
    private ConnectionManager connectionManager = null;
    /**
     * The ManagedConnectionFactory associated to this ConnectionFactory.
     */
    private JdoManagedConnectionFactory managedConnectionFactory;

    /**
     * Constructs a JdoConnectionFactory. Associates a ConnectionManager with
     * it, or creates a default one (if null).
     * @param fmcf	The ManagedConnectionFactory to be associated with this
     * 				ConnectionFactory.
     * @param cm	The ConnectionManager to be associated with this
     * 				ConnectionFactory. It may be null.
     */
    JdoConnectionFactory(Logger logger,
						 JdoManagedConnectionFactory fmcf,
						 ConnectionManager cm) throws ResourceException {
        this.logger = logger;
        if (Debug.ON)
            logger.log(BasicLevel.DEBUG,
                       "Constructs a new JdoConnectionFactory.");
        managedConnectionFactory = fmcf;
        setConnectionManager(cm);
    }

    /**
     * Creates a new JdoConnection.
     */
    JdoConnection createConnection() throws ResourceException {
        JdoConnection res = new JdoConnection(logger, this);
        if (Debug.ON)
            logger.log(BasicLevel.DEBUG, "New JdoConnection created: " + res);
        return res;
    }

    private void initialize() {
        if (!started) {
            try {
                managedConnectionFactory.start();
                started = true;
            } catch (ResourceException e) {
                throw new JDOException("JDO Connector: cannot initialize the ManagedConnectionFactory.", e);
            }
        }
    }

    public void setConnectionManager(ConnectionManager connectionManager) {
        if (connectionManager != null) {
            logger.log(BasicLevel.DEBUG,
                    "Linked the ConnectionFactory to the ConnectionManager: "
                    + connectionManager);
        }
		this.connectionManager = connectionManager;
    }

    // IMPLEMENTATION OF METHODS FROM THE (cci)ConnectionFactory INTERFACE

    /**
     * Connection allocation is delegated to the ConnectionManager assigned to
     * this factory.
     */
    public Connection getConnection() throws ResourceException {
        if (!started) {
            managedConnectionFactory.start();
            started = true;
        }
		return (Connection)
				connectionManager.allocateConnection(
                managedConnectionFactory, null);
    }

    /**
     * Connection allocation is delegated to the ConnectionManager assigned to
     * this factory.
     * @param spec	The connection is not taken into account within JDO.
     */
    public Connection getConnection(ConnectionSpec spec) throws ResourceException {
        if (!started) {
            managedConnectionFactory.start();
            started = true;
        }
		return (Connection)
				connectionManager.allocateConnection(
                managedConnectionFactory, null);
    }

    /**
     * Assigns a JNDI Reference to the adapter.
     * @param reference	The JNDI Reference to be assigned.
     */
    public void setReference(Reference reference) {
        jndiReference = reference;
    }

    /**
     * Returns the JNDI Reference.
     * @return	The JNDI Reference.
     */
    public Reference getReference() throws NamingException {
        return jndiReference;
    }

    /**
     * No support for record.
     */
    public RecordFactory getRecordFactory() throws ResourceException {
        throw new NotSupportedException("JDO Connector: no support for record.");
    }

    /**
     * The JdoConnectionFactory manages its metadata on its own.
     */
    public ResourceAdapterMetaData getMetaData() throws ResourceException {
        return this;
    }

    // IMPLEMENTATION OF METHODS FROM THE (cci)ResourceAdapaterMetaData INTERFACE

    /**
     * Gives access to the adapter version metadata for the JDO.
     * @return	The JDO adapter version.
     */
    public String getAdapterVersion() {
        return JDOADAPTERVERSION;
    }

    /**
     * Gives access to the adapter vendor name metadata for the JDO.
     * @return	The JDO adapter vendor name.
     */
    public String getAdapterVendorName() {
        return JDOADAPTERVENDORNAME;
    }

    /**
     * Gives access to the adapter name metadata for the JDO.
     * @return	The JDO adapter name.
     */
    public String getAdapterName() {
        return JDOADAPTERNAME;
    }

    /**
     * Gives access to a short description of the JDO adapter.
     * @return	The JDO short description.
     */
    public String getAdapterShortDescription() {
        return JDOADAPTERSHORTDESCRIPTION;
    }

    /**
     * Gives access to the specification version of the JDO.
     * @return	The JDO specification version.
     */
    public String getSpecVersion() {
        return JDOSPECVERSION;
    }

    /**
     * No interaction is supported by the JDO adapter.
     */
    public String[] getInteractionSpecsSupported() {
        return new String[0];
    }

    /**
     * No support for record.
     */
    public boolean supportsExecuteWithInputAndOutputRecord() {
        return false;
    }

    /**
     * No support for record.
     */
    public boolean supportsExecuteWithInputRecordOnly() {
        return false;
    }

    /**
     * Yes, it does support local transaction demarcation.
     */
    public boolean supportsLocalTransactionDemarcation() {
        return true;
    }

    // IMPLEMENTATION OF METHODS FROM THE PersistenceManagerFactory INTERFACE

    public PersistenceManager getPersistenceManager() {
        try {
			PersistenceManager pm = (PersistenceManager) getConnection();
			pm.currentTransaction();
            return pm;
        } catch (ResourceException e) {
            throw new JDOException("JDO Connector: problem while getting a PersistenceManager.", e);
        }
    }

    public PersistenceManager getPersistenceManager(String s, String s1) {
        try {
			PersistenceManager pm = (PersistenceManager) getConnection();
			pm.currentTransaction();
            return pm;
        } catch (ResourceException e) {
            throw new JDOException("JDO Connector: problem while getting a PersistenceManager.", e);
        }
    }

    public void setConnectionUserName(String s) {
        initialize();
        managedConnectionFactory.setConnectionUserName(s);
    }

    public String getConnectionUserName() {
        initialize();
        return managedConnectionFactory.getConnectionUserName();
    }

    public void setConnectionPassword(String s) {
        initialize();
        managedConnectionFactory.setConnectionPassword(s);
    }

    public void setConnectionURL(String s) {
        initialize();
        managedConnectionFactory.setConnectionURL(s);
    }

    public String getConnectionURL() {
        initialize();
        return managedConnectionFactory.getConnectionURL();
    }

    public void setConnectionDriverName(String s) {
        initialize();
        managedConnectionFactory.setConnectionDriverName(s);
    }

    public String getConnectionDriverName() {
        initialize();
        return managedConnectionFactory.getConnectionDriverName();
    }

    public void setConnectionFactoryName(String s) {
        initialize();
        managedConnectionFactory.setConnectionFactoryName(s);
    }

    public String getConnectionFactoryName() {
        initialize();
        return managedConnectionFactory.getConnectionFactoryName();
    }

    public void setConnectionFactory(Object o) {
        initialize();
        managedConnectionFactory.setConnectionFactory(o);
    }

    public Object getConnectionFactory() {
        initialize();
        return managedConnectionFactory.getConnectionFactory();
    }

    public void setConnectionFactory2Name(String s) {
        initialize();
        managedConnectionFactory.setConnectionFactory2Name(s);
    }

    public String getConnectionFactory2Name() {
        initialize();
        return managedConnectionFactory.getConnectionFactory2Name();
    }

    public void setConnectionFactory2(Object o) {
        initialize();
        managedConnectionFactory.setConnectionFactory2(o);
    }

    public Object getConnectionFactory2() {
        initialize();
        return managedConnectionFactory.getConnectionFactory2();
    }

    public void setMultithreaded(boolean b) {
        initialize();
        managedConnectionFactory.setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        initialize();
        return managedConnectionFactory.getMultithreaded();
    }

    public void setOptimistic(boolean b) {
        initialize();
        managedConnectionFactory.setOptimistic(b);
    }

    public boolean getOptimistic() {
        initialize();
        return managedConnectionFactory.getOptimistic();
    }

    public void setRetainValues(boolean b) {
        initialize();
        managedConnectionFactory.setRetainValues(b);
    }

    public boolean getRetainValues() {
        initialize();
        return managedConnectionFactory.getRetainValues();
    }

    public void setRestoreValues(boolean b) {
        initialize();
        managedConnectionFactory.setRestoreValues(b);
    }

    public boolean getRestoreValues() {
        initialize();
        return managedConnectionFactory.getRestoreValues();
    }

    public void setNontransactionalRead(boolean b) {
        initialize();
        managedConnectionFactory.setNontransactionalRead(b);
    }

    public boolean getNontransactionalRead() {
        initialize();
        return managedConnectionFactory.getNontransactionalRead();
    }

    public void setNontransactionalWrite(boolean b) {
        initialize();
        managedConnectionFactory.setNontransactionalWrite(b);
    }

    public boolean getNontransactionalWrite() {
        initialize();
        return managedConnectionFactory.getNontransactionalWrite();
    }

    public void setIgnoreCache(boolean b) {
        initialize();
        managedConnectionFactory.setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        initialize();
        return managedConnectionFactory.getIgnoreCache();
    }

    public int getMaxPool() {
        initialize();
        return managedConnectionFactory.getMaxPool();
    }

    public void setMaxPool(int i) {
        initialize();
        managedConnectionFactory.setMaxPool(i);
    }

    public int getMinPool() {
        initialize();
        return managedConnectionFactory.getMinPool();
    }

    public void setMinPool(int i) {
        initialize();
        managedConnectionFactory.setMinPool(i);
    }

    public int getMsWait() {
        initialize();
        return managedConnectionFactory.getMsWait();
    }

    public void setMsWait(int i) {
        initialize();
        managedConnectionFactory.setMsWait(i);
    }

    public Properties getProperties() {
        initialize();
        return managedConnectionFactory.getProperties();
    }

    public Collection supportedOptions() {
        initialize();
        return managedConnectionFactory.supportedOptions();
    }
}
