/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.speedo.jca;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Loggable;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.api.SpeedoProperties;

import java.io.PrintWriter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.FileNotFoundException;
import java.util.Set;
import java.util.Properties;
import java.util.Collection;
import javax.resource.ResourceException;
import javax.resource.cci.ConnectionFactory;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.security.auth.Subject;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.JDOHelper;
import javax.jdo.PersistenceManager;
import javax.transaction.TransactionManager;
import javax.naming.InitialContext;

/**
 * @author	P. Dechamboux
 */
public class JdoManagedConnectionFactory
        implements ManagedConnectionFactory,
        JdoAttributeController, PersistenceManagerFactory {

    public final static String MANAGED_PROPERTY =
            "org.objectweb.perseus.connector.ra.jdo.managed";

    private final static String[] DEFAULT_JNDI_NAMES = {
        "javax.transaction.UserTransaction", 	//JOnAS
		"javax.transaction.TransactionManager", //WebLogic
		"java:/TransactionManager", 			//JBoss
        "java:comp/UserTransaction" 			//Orion
    };

    /**
     * The logger into which traces about JdoManagedConnectionFactory are
     * produced.
     */
    private Logger logger;
    /**
     * It is assumed that only one ConnectionFactory is actually created by
     * a ManagedConnectionFactory.
     */
    private JdoConnectionFactory connectionFactory;
    /**
     * The factory for managing JDO XA resources.
     */
    private JdoXAResourceFactory xaResourceFactory = null;
    /**
     * For creating all loggers related to JDO and its adapter.
     */
    private LoggerFactory loggerFactory = null;
    /**
     * The prefix added to the name of each logger created for the purpose of
     * the JDO or its adapter.
     */
    private String loggerBaseName = "";
    /**
     * The factory for managing JDO transaction contexts.
     */
    private PersistenceManagerFactory txContextFactory = null;
    /**
     * The name of the property file of the PersistenceManagerFactory associated
     * with this JDO connector.
     */
    private String propertiesFileName = null;
    /**
     * The factory for managing JDO transaction contexts.
     */
    private ConnectionManager connectionManager = null;

    private PrintWriter printWriter = null;

    private boolean started = false;

    /**
     * the JNDI Name of the transaction manager
     */
    private String tmName = null;

    private TransactionManager tm = null;


    /**
     * Lookup in JNDI the transaction manager under the name specified in
     * parameter. The 'tm' variable is assigned if the transaction manager is
     * found.
     * @param name is the jndi name
     * @return true if an instance implementing
     * javax.transaction.TransactionManager is availlable in JNDI, otherwise
     * false.
     */
    private boolean lookupTM(String name) {
        try {
            Object o = new InitialContext().lookup(name);
            if (o == null) {
                logger.log(BasicLevel.WARN,
                        "JNDI retrieves a null value for the name '" + name + "'.");
            }
            if (!(o instanceof TransactionManager)) {
                logger.log(BasicLevel.WARN,
                        "JNDI retrieves an object which is not a javax.transaction.TransactionManager (JNDI name: " + name + "): " + o);
                return false;
            }
            tm = (TransactionManager) o;
            logger.log(BasicLevel.INFO, "The TransactionManager was found in JNDI with the name '" + name + "'.");
            return true;
        } catch (Exception e) {
            logger.log(BasicLevel.WARN,
                    "Error when lookup the transaction manager in JNDI with the name '"
                    + name + "'", e);
            return false;
        }
    }

    /**
     * Starts this JdoManagedConnectionFactory.
     */
    public void start() throws ResourceException {
        //Logging initialisation
        if (loggerFactory == null) {
            if (Monolog.loggerFactory == null) {
                loggerFactory = Monolog.init();
            } else {
                loggerFactory = Monolog.loggerFactory;
            }
        }
        if (logger == null) {
            logger = loggerFactory.getLogger("org.objectweb.speedo.jca");
        }

        //Load properties file of the JDO driver
        Properties p = new Properties();
        if (txContextFactory == null) {
            if (propertiesFileName == null) {
                throw new ResourceException("No name provided for the properties file of the associated PersistenceManagerFactory");
            }
            InputStream is = getClass().getClassLoader()
                        .getResourceAsStream(propertiesFileName);
            if (is == null) {
                File f = new File(propertiesFileName);
                try {
                    if (f.exists()) {
                        is = new FileInputStream(propertiesFileName);
                        logger.log(BasicLevel.DEBUG, "Properties file '"
                                + propertiesFileName + "'  found in the file system");
                    }
                } catch (FileNotFoundException e) {
                } finally {
                    if (is == null) {
                        throw new ResourceException(
                                "Unable to load properties file: "
                                + propertiesFileName);
                    }
                }
            } else {
                logger.log(BasicLevel.DEBUG, "Properties file '"
                        + propertiesFileName + "'  found in the classpath");
            }
            try {
                p.load(is);
            } catch (IOException e) {
                throw new ResourceException("Unable to load properties file: "
                        + propertiesFileName);
            }
            p.put(MANAGED_PROPERTY, "true");
            logger.log(BasicLevel.DEBUG, "Properties loaded:" + p);

            /**
             * Try to find the TransactionManager with:
             * - the name specified in the RA configuration xml file
             * - the name specified the properties of the JDO driver
             * - names used in some application server
             */
            if (tm == null) {
                if (tmName != null && lookupTM(tmName)) {
                    p.setProperty(SpeedoProperties.TM_NAME, tmName);
                } else {
                    //lookup in the properties of JDO driver if the TM name is specified
                    String tmName2 = p.getProperty(SpeedoProperties.TM_NAME);
                    if (tmName2 == null || !lookupTM(tmName2)) {
                        logger.log(BasicLevel.DEBUG,
                                "Try to find the transaction manager with default JNDI names.");
                        for(int i=0;
                            i<DEFAULT_JNDI_NAMES.length && !lookupTM(DEFAULT_JNDI_NAMES[i]);
                            i++) {
                            p.setProperty(SpeedoProperties.TM_NAME, DEFAULT_JNDI_NAMES[i]);
                        }
                    }
                }
            }
            if (tm == null) {
                throw new ResourceException(
                        "A javax.transaction.TransactionManager instance is required," +
                        " in order to register the JDO driver as a Synchronization on transaction"
                        + (tmName == null
                            ? "(No JNDI name specified)"
                            : "(Bad JNDI Name)"));
            } else {
                logger.log(BasicLevel.DEBUG, "Transaction manager found: " + tm);
            }
            connectionFactory = new JdoConnectionFactory(logger, this, connectionManager);
            logger.log(BasicLevel.INFO, "ConnectionManager allocated");
            //fetch the real PersistenceManagerFactory
            txContextFactory = JDOHelper.getPersistenceManagerFactory(p);
            logger.log(BasicLevel.INFO, "JdoManagedConnectionFactory started");
        }
        started = true;
    }

    protected void finalize() throws Throwable {
        stop();
        super.finalize();
    }

    /**
     * Stops this JdoManagedConnectionFactory.
     */
    public void stop() throws ResourceException {
        txContextFactory = null;
    }


    /**
     * Delegates the creation of a Connection to the ConnectionFactory.
     */
    public Object createConnection() throws ResourceException {
        return connectionFactory.createConnection();
    }

    /**
     * Retrieves the JdoTxContextFactory.
     */
    PersistenceManagerFactory getTxContextFactory() {
        if (Debug.ON)
            logger.log(BasicLevel.DEBUG, "TxContextFactory is: " + txContextFactory);
        return txContextFactory;
    }

    /**
     * Retrieves the ConnectionFactory.
     */
    ConnectionFactory getJcaConnectionFactory() {
        return connectionFactory;
    }

    /**
     * Retrieves the XAResourceFactory. If null, creates it. This means that no
     * object related to XAResource management is created if not executing in
     * XA-aware runtime environment.
     */
    JdoXAResourceFactory getXAResourceFactory() {
        if (xaResourceFactory == null) {
            xaResourceFactory = new JdoXAResourceFactory(logger, this);
        }
        return xaResourceFactory;
    }

    public TransactionManager getTransactionManager() {
        return tm;
    }

    // IMPLEMENTATION OF METHODS FROM THE JdoAttributeController INTERFACE

    /**
     * Gives access to the name of the property file for initializing the
     * underlying JDO implementation.
     * @return  The name of the property file.
     */
    public String getPropertyFile() {
        return propertiesFileName;
    }

    /**
     * Assigns to this JDO connector the name of the property file for
     * initializing the underlying JDO implementation.
     * @param pf  The name of the property file.
     */
    public void setPropertyFile(String pf) {
        propertiesFileName = pf;
    }

    public String getTransactionManagerJNDIName() {
        return tmName;
    }

    public void setTransactionManagerJNDIName(String jndiname) throws ResourceException {
        tmName = jndiname;
    }

    public void setTransactionManager(TransactionManager tm) {
        this.tm = tm;
    }

    // IMPLEMENTATION OF METHODS FROM THE (cci)ManagedConnectionFactory INTERFACE

    /**
     * Creates a JdoConnectionFactory; yields the existing one if any.
     * @param cm	The ConnectionManager to be used by the created
     * 				ConnectionFactory (may be null).
     */
    public Object createConnectionFactory(ConnectionManager cm)
            throws ResourceException {
        if (!started) {
            start();
        }
        connectionFactory.setConnectionManager(cm);
        return connectionFactory;
    }

    /**
     * Creates a JdoConnectionFactory; yields the existing one if any.
     */
    public Object createConnectionFactory() throws ResourceException {
        return createConnectionFactory(null);
    }

    /**
     * Creates a new JdoManagedConnection.
     */
    public ManagedConnection createManagedConnection(
            Subject subject,
            ConnectionRequestInfo info) throws ResourceException {
        if (logger == null || txContextFactory == null) {
            start();
        }
        JdoManagedConnection jmc = new JdoManagedConnection(logger, this);
        return jmc;
    }

    /**
     * No matching rules supported. Always yields the first element of the set
     * if any.
     */
    public ManagedConnection matchManagedConnections(
            Set set,
            Subject subject,
            ConnectionRequestInfo info) throws ResourceException {
        if (set.size() == 0)
            return null;
        JdoManagedConnection jmc = (JdoManagedConnection)
                set.iterator().next();
        return jmc;
    }

    /**
     * If he given PrintWrtier is a Loggable implementation then the inner
     * logger and the inner loggerFactory are used. Otherwise the a basic Logger
     * implementation is used over the specified PrintWriter.
     */
    public void setLogWriter(PrintWriter writer) throws ResourceException {
        if (logger == null) {
            if (writer instanceof Loggable) {
                logger = ((Loggable) writer).getLogger();
                loggerFactory = ((Loggable) writer).getLoggerFactory();
            } else {
                LoggerImpl li = new LoggerImpl(writer);
                logger = li;
                loggerFactory = li;
            }
        }
        printWriter = writer;
    }

    /**
     * Retrieves the printwriter used for the logging.
     */
    public PrintWriter getLogWriter() throws ResourceException {
        return printWriter;
    }

    // IMPLEMENTATION OF METHODS FROM THE JdoAttributeController INTERFACE

    /**
     * Gives access to the base name used as a prefix when registering new
     * loggers to the LoggerFactory.
     */
    public String getLoggerBaseName() {
        return loggerBaseName;
    }

    /**
     * Changes the base name used as a prefix when registering new
     * loggers to the LoggerFactory.
     */
    public void setLoggerBaseName(String basename) {
        loggerBaseName = basename;
    }

    /**
     * Gives access to the ConnectionManager associated with a JDO adapter.
     */
    public ConnectionManager getConnectionManager() {
        return connectionManager;
    }

    /**
     * Changes the ConnectionManager associated with a JDO adapter.
     */
    public void setConnectionManager(ConnectionManager cm) {
        connectionManager = cm;
    }

    // IMPLEMENTATION OF METHODS FROM THE PersistenceManagerFactory INTERFACE

    public PersistenceManager getPersistenceManager() {
        return null;
    }

    public PersistenceManager getPersistenceManager(String s, String s1) {
        return null;
    }

    public void setConnectionUserName(String s) {
        txContextFactory.setConnectionUserName(s);
    }

    public String getConnectionUserName() {
        return txContextFactory.getConnectionUserName();
    }

    public void setConnectionPassword(String s) {
        txContextFactory.setConnectionPassword(s);
    }

    public void setConnectionURL(String s) {
        txContextFactory.setConnectionURL(s);
    }

    public String getConnectionURL() {
        return txContextFactory.getConnectionURL();
    }

    public void setConnectionDriverName(String s) {
        txContextFactory.setConnectionDriverName(s);
    }

    public String getConnectionDriverName() {
        return txContextFactory.getConnectionDriverName();
    }

    public void setConnectionFactoryName(String s) {
        txContextFactory.setConnectionFactoryName(s);
    }

    public String getConnectionFactoryName() {
        return txContextFactory.getConnectionFactoryName();
    }

    public void setConnectionFactory(Object o) {
        txContextFactory.setConnectionFactory(o);
    }

    public Object getConnectionFactory() {
        return txContextFactory.getConnectionFactory();
    }

    public void setConnectionFactory2Name(String s) {
        txContextFactory.setConnectionFactory2Name(s);
    }

    public String getConnectionFactory2Name() {
        return txContextFactory.getConnectionFactory2Name();
    }

    public void setConnectionFactory2(Object o) {
        txContextFactory.setConnectionFactory2(o);
    }

    public Object getConnectionFactory2() {
        return txContextFactory.getConnectionFactory2();
    }

    public void setMultithreaded(boolean b) {
        txContextFactory.setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        return txContextFactory.getMultithreaded();
    }

    public void setOptimistic(boolean b) {
        txContextFactory.setOptimistic(b);
    }

    public boolean getOptimistic() {
        return txContextFactory.getOptimistic();
    }

    public void setRetainValues(boolean b) {
        txContextFactory.setRetainValues(b);
    }

    public boolean getRetainValues() {
        return txContextFactory.getRetainValues();
    }

    public void setRestoreValues(boolean b) {
        txContextFactory.setRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return txContextFactory.getRestoreValues();
    }

    public void setNontransactionalRead(boolean b) {
        txContextFactory.setNontransactionalRead(b);
    }

    public boolean getNontransactionalRead() {
        return txContextFactory.getNontransactionalRead();
    }

    public void setNontransactionalWrite(boolean b) {
        txContextFactory.setNontransactionalWrite(b);
    }

    public boolean getNontransactionalWrite() {
        return txContextFactory.getNontransactionalWrite();
    }

    public void setIgnoreCache(boolean b) {
        txContextFactory.setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        return txContextFactory.getIgnoreCache();
    }

    public int getMaxPool() {
        return txContextFactory.getMaxPool();
    }

    public void setMaxPool(int i) {
        txContextFactory.setMaxPool(i);
    }

    public int getMinPool() {
        return txContextFactory.getMinPool();
    }

    public void setMinPool(int i) {
        txContextFactory.setMinPool(i);
    }

    public int getMsWait() {
        return txContextFactory.getMsWait();
    }

    public void setMsWait(int i) {
        txContextFactory.setMsWait(i);
    }

    public Properties getProperties() {
        return txContextFactory.getProperties();
    }

    public Collection supportedOptions() {
        return txContextFactory.supportedOptions();
    }
}