/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.naming.api;

import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.generation.jorm.JormMIMappingBuilder;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.HelperClassMapping;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.util.monolog.api.Logger;

import java.util.Collection;
import java.util.Map;

/**
 *
 * @author S.Chassande-Barrioz
 */
public interface NamingManager {

    String SEP = "/";

    /**
     * Indicates if the naming manager is able to manage the identifier of a
     * persistent class.
     * @param sc The speedo meta object representing the persistent class
     */
    boolean canManage(SpeedoClass sc);

    boolean canProvidePBinder(Object hints, ClassLoader classLoader);

    boolean canProvidePNamingContext(Object hints, ClassLoader classLoader);

    PBinder getPBinder(String className, String hints, ClassLoader classLoader, byte mappingStructureRule) throws PException;

    PNamingContext getPNamingContext(String className, String hints, ClassLoader classLoader, Map binders) throws PException;

    PName decode(PNameCoder pnc, Object oid, Class clazz, JormFactory jf) throws PException;

    Object encode(PName pn) throws PException;

    boolean supportPNamingcontext();

    /**
     * Fill the name def of an identifier or a reference
     * @param nd is the name def to fill
     * @param isIdentifier indicates if the name represents an identifier (true)
     * or a reference (false).
     * @param isInGenClass indicates if the name is defined in a generic class
     * (true) of in a class (false).
     * @param ref meta object which the name def must be
     * defined. This value is used only in the case of isIdentifier == false and
     * isInGenClass == false.
     * @param tsc is the speedo meta object representing the referenced class.
     * This value is used only in the case of isIdentifier == false.
     * @param mo is the jorm meta object hosting the name def and on which the
     * eventual hidden field will be created.
     * @param hcm if the mapping structure hosting the mapping of the reference.
     * This value must be ClassMapping or GenClassMapping instance.
     * @param mb is the mapping builder permitting the creation of the mapping
     * part.
     * @param createdMOs is a result paramter. This collection must be fill with
     * the created Jorm Meta objects representing  a class or a composite name.
     * Here only the new used composite name will be added.
     * @throws SpeedoException if the speedo meta information is not completly
     * defined.
     */
    void fillNameDef(MIBuilderHelper mibh,
                     Manager manager,
                     NameDef nd,
                     SpeedoClass tsc,
                     SpeedoClass ssc,
                     MetaObject mo,
                     Reference ref,
                     HelperClassMapping hcm,
                     JormMIMappingBuilder mb,
                     boolean isIdentifier,
                     boolean isInGenClass,
                     boolean createField,
                     Collection createdMOs) throws SpeedoException, PException;

    String getJormConfig(NameDef nd,
                         SpeedoClass targetClass,
                         MetaObject sourceMO);

    String getPNameHints(SpeedoClass sc, NameDef nd);

    String getGCPNameHints(SpeedoClass sc, NameDef nd);

    void setPMapper(PMapper mapper);

    void setLogger(Logger logger);

    interface PBinderProvider {
        PBinder findPBinder(String bcn, String clName, ClassLoader classLoader)
                throws PException;
    }

    NamingField[] getNamingfields(SpeedoClass sc) throws PException;

    class NamingField {

        private String userFieldName = null;
        private Class userFieldType = null;
        private String namingFieldName = null;
        private PType namingFieldType = null;

        public NamingField(String userFieldName,
                           Class userFieldType,
                           String namingFieldName,
                           PType namingFieldType) throws PException {
            this.userFieldName = userFieldName;
            this.userFieldType = userFieldType;
            this.namingFieldName = namingFieldName;
            this.namingFieldType = namingFieldType;
        }

        public String getUserFieldName() {
            return userFieldName;
        }

        public Class getUserFieldType() {
            return userFieldType;
        }

        public String getNamingFieldName() {
            return namingFieldName;
        }

        public PType getNamingFieldType() {
            return namingFieldType;
        }

        public String getPNameGetter() {
            switch (namingFieldType.getTypeCode()) {
            case PType.TYPECODE_CHAR:
                return "pngetCharField";
            case PType.TYPECODE_OBJCHAR:
                return "pngetOcharField";
            case PType.TYPECODE_BYTE:
                return "pngetByteField";
            case PType.TYPECODE_OBJBYTE:
                return "pngetObyteField";
            case PType.TYPECODE_SHORT:
                return "pngetShortField";
            case PType.TYPECODE_OBJSHORT:
                return "pngetOshortField";
            case PType.TYPECODE_INT:
                return "pngetIntField";
            case PType.TYPECODE_OBJINT:
                return "pngetOintField";
            case PType.TYPECODE_LONG:
                return "pngetLongField";
            case PType.TYPECODE_OBJLONG:
                return "pngetOlongField";
            case PType.TYPECODE_STRING:
                return "pngetStringField";
            case PType.TYPECODE_DATE:
                return "pngetDateField";
            case PType.TYPECODE_CHARARRAY:
                return "pngetCharArrayField";
            case PType.TYPECODE_BYTEARRAY:
                return "pngetByteArrayField";
            default :
                return null;
            }
        }
    }
}