/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime;

import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.lib.BasicWorkingSet;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.Transaction;
import org.objectweb.speedo.api.WorkingSetLifeCycle;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.TransactionListener;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.ComponentIdentity;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;

import javax.jdo.JDOException;
import javax.jdo.PersistenceManager;
import javax.jdo.JDOFatalException;
import javax.transaction.Synchronization;
import javax.transaction.Status;
import javax.transaction.TransactionManager;
import javax.resource.ResourceException;
import javax.naming.InitialContext;
import java.sql.SQLException;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoTransaction
        extends BasicWorkingSet
        implements Transaction, LifeCycleController {

    public final static String PROXY_MANAGER_BINDING = "proxy-manager";
    public final static String PMAPPER_BINDING = "mapper";
    public final static String TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING
            = "transactional-persistence-manager";
    public final static String COMPONENT_IDENTITY_BINDING = "component-identity";
	public static TransactionListener txListener = null;

    private ProxyManager pm = null;
	private PMapper mapper = null;
    private TransactionalPersistenceManager tpm = null;
    private Transaction _this;

    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;
    private boolean retainValues;
    private boolean restoreValues;
    private boolean optimistic;
    private Synchronization synchronization = null;
    private Object dataAccess = null;
    private boolean managedEnv = false;

    /**
     * It activates the working set. This is used to delimit the begining of
     * the working set.
     */
    public void active() throws PersistenceException {
        status = WorkingSetLifeCycle.getNextStatus(
                status, WorkingSetLifeCycle.ACTIVE_ACTION);
    }

	// IMPLEMENTATION OF THE Transaction INTERFACE //
	//-----------------------------------------------------//

	public JDOFatalException rollBackOnInternalError(Exception _e) {
		Exception ie = ExceptionHelper.getNested(_e);
		if (isActive()) {
			if (managedEnv) {
				logger.log(BasicLevel.INFO, ".");
				String tmName = pm.getPersistenceManagerFactory().getProperties()
						.getProperty(SpeedoProperties.TM_NAME);
				if (tmName == null) {
					return new JDOFatalException("No transaction manager jndi name found in initialisation properties");
				}
				try {
					Object o = new InitialContext().lookup(tmName);
					if (o == null) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves a null transaction manager for the name '" + tmName + "'.";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					if (!(o instanceof TransactionManager)) {
						String msg = "The transaction must be marked as rollbackOnly: JNDI retrieves an object which is not a javax.transaction.TransactionManager (JNDI name: " + tmName + "): " + o;
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					javax.transaction.Transaction t = ((TransactionManager) o).getTransaction();
					if (t == null) {
						String msg = "Impossible to rollback an unexisting transaction (TM.getTransaction()=null)";
						logger.log(BasicLevel.ERROR, msg);
						return new JDOFatalException(msg);
					}
					t.setRollbackOnly();
				} catch (Exception e) {
					String msg = "The transaction must be marked as rollbackOnly: Error when lookup the transaction manager in JNDI with the name '"
							+ tmName + "'";
					logger.log(BasicLevel.ERROR, msg, e);
					return new JDOFatalException(msg);
				}
			} else {
				rollback();
			}
			return new JDOFatalException("The current transaction has been rolledback, please retry", ie);
		} else {
			return new JDOFatalException("The current woring set occrurs an error, please retry", ie);
		}
	}

    public boolean isManagedEnv() {
        return managedEnv;
    }

    // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
    //-----------------------------------------------------//

    public String getFcState() {
        return null;
    }

    public void startFc() {
        managedEnv = pm.getPersistenceManagerFactory().getProperties()
            .getProperty(SpeedoProperties.MANAGED, "")
            .equals("true");
    }

    public void stopFc() {
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String c) {
        if (PROXY_MANAGER_BINDING.equals(c))
            return pm;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            return tpm;
        else if (PMAPPER_BINDING.equals(c))
            return mapper;
        else
            return super.getFcBindings(c);
    }

    public void addFcBinding(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = (ProxyManager) s;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = (TransactionalPersistenceManager) s;
        else if (PMAPPER_BINDING.equals(c))
            mapper = (PMapper) s;
        else if (COMPONENT_IDENTITY_BINDING.equals(c)) {
            _this = (Transaction) ((ComponentIdentity) s).getFcInterface("transaction");
        } else
            super.addFcBinding(c, s);
    }

    public void removeFcBinding(String c, Object s) {
        if (PROXY_MANAGER_BINDING.equals(c))
            pm = null;
        else if (TRANSACTIONAL_PERSISTENCE_MANAGER_BINDING.equals(c))
            tpm = null;
        else if (PMAPPER_BINDING.equals(c))
            mapper = null;
        else
            super.removeFcBinding(c, s);
    }


	// IMPLEMENTATION OF THE SpeedoConnectionHolder INTERFACE //
	//--------------------------------------------------------//

	public void releaseCH() throws PException {
		if (dataAccess != null && managedEnv && isActive()) {
			logger.log(BasicLevel.DEBUG, "Release and close the connection " + dataAccess);
			mapper.closeConnection(dataAccess);
			dataAccess = null;
		}
	}

	/**
	 * Retrieves a connection to access data on the support.
	 * @throws PException if it is not possible to find a connection
	 */
	public Object getCHConnection() throws PException {
		if (dataAccess == null) {
			Object cs = pm.getConnectionSpec();
			try {
				if (cs == null) {
					dataAccess = mapper.getConnection();
				} else {
					dataAccess = mapper.getConnection(cs);
				}
			} catch (PException e) {
				throw new PException(e, "Impossible to fetch a data access");
			}
			logger.log(BasicLevel.DEBUG, "Allocate a connection: " + dataAccess);
			if (isActive() && !managedEnv) {
				logger.log(BasicLevel.DEBUG, "Start a local transaction");
				//Open a local transaction
				if (dataAccess instanceof java.sql.Connection) {
					try {
						((java.sql.Connection) dataAccess).setAutoCommit(false);
					} catch (SQLException e) {
						throw new JDOException("", ExceptionHelper.getNested(e));
					}
				} else if (dataAccess instanceof javax.resource.cci.Connection) {
					try {
						((javax.resource.cci.Connection) dataAccess)
								.getLocalTransaction().begin();
					} catch (ResourceException e) {
						throw new JDOException("", ExceptionHelper.getNested(e));
					}
				}
			}
		}
		return dataAccess;
	}

	public void commitCH() throws PException {
		if (dataAccess != null && !managedEnv && isActive()) {
			logger.log(BasicLevel.DEBUG, "Commit the local transaction on the connection: " + dataAccess);
			if (dataAccess instanceof java.sql.Connection) {
				try {
					((java.sql.Connection) dataAccess).commit();
					((java.sql.Connection) dataAccess).setAutoCommit(true);
				} catch (SQLException e) {
					throw new JDOException("", ExceptionHelper.getNested(e));
				}
			} else if (dataAccess instanceof javax.resource.cci.Connection) {
				try {
					((javax.resource.cci.Connection) dataAccess)
							.getLocalTransaction().rollback();
				} catch (ResourceException re) {
					throw new JDOException("", ExceptionHelper.getNested(re));
				}
			}
		}
	}

	public void rollbackCH() throws PException {
		if (dataAccess != null && !managedEnv && isActive()) {
			logger.log(BasicLevel.DEBUG, "Rollback the local transaction on the connection: " + dataAccess);
			if (dataAccess instanceof java.sql.Connection) {
				try {
					((java.sql.Connection) dataAccess).rollback();
					((java.sql.Connection) dataAccess).setAutoCommit(true);
				} catch (SQLException e) {
					throw new JDOException("", ExceptionHelper.getNested(e));
				}
			} else if (dataAccess instanceof javax.resource.cci.Connection) {
				try {
					((javax.resource.cci.Connection) dataAccess)
							.getLocalTransaction().rollback();
				} catch (ResourceException re) {
					throw new JDOException("", ExceptionHelper.getNested(re));
				}
			}
		}
	}

	public void closeCH() throws PException {
		if (dataAccess != null) {
			logger.log(BasicLevel.DEBUG, "Close the connection: " + dataAccess);
			mapper.closeConnection(dataAccess);
			dataAccess = null;
		}
	}

    // IMPLEMENTATION OF THE javax.jdo.Transactionn INTERFACE //
    //--------------------------------------------------------//

    public void begin() {
		logger.log(BasicLevel.DEBUG, "Begin the transaction");
		if (txListener != null) {
			txListener.transactionBegun(this);
		}
        try {
            tpm.begin(_this);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the begin of the transaction:", ie);
            throw new JDOException("", ie);
        }
    }

    public void commit() {
        logger.log(BasicLevel.DEBUG, "Commit the transaction, working set size: " + ces.size());
        if (synchronization != null)
            synchronization.beforeCompletion();
        boolean validated;
        try {
            validated = tpm.prepare(_this, true);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR, "Roll back a transaction: ", ie);
            rollbackPrepared();
            logger.log(BasicLevel.ERROR,
                    "Error during the prepare action on the transaction:", ie);
            throw new JDOFatalException("Rollback due to a prepare problem", ie);
        }
        if (validated) {
			int size = ces.size();
            try {
                tpm.commit(_this);
            } catch (PersistenceException e) {
                Exception ie = ExceptionHelper.getNested(e);
                logger.log(BasicLevel.ERROR,
                        "Error during the commit of the transaction:", ie);
                throw new JDOException("", ie);
            }
            if (synchronization != null)
                synchronization.afterCompletion(Status.STATUS_COMMITTED);
			if (txListener != null) {
				txListener.transactionCommitted(this, size);
			}
        } else {
            rollbackPrepared();
			throw new JDOFatalException("Rollback due to a concurrency problem");
        }
    }

    private void rollbackPrepared() {
		int size = ces.size();
        try {
            tpm.rollback(_this);
        } catch (PersistenceException e) {
            Exception ie = ExceptionHelper.getNested(e);
            logger.log(BasicLevel.ERROR,
                    "Error during the rollback of the transaction:", ie);
            throw new JDOException("", ie);
        }
		if (txListener != null) {
			txListener.transactionAborted(this, size);
		}
        if (synchronization != null)
            synchronization.afterCompletion(Status.STATUS_ROLLEDBACK);
    }
	
    public void rollback() {
		if (status != TransactionalPersistenceManager.CTX_ACTIVE_TRANSACTIONAL) {
			return;
		}
        logger.log(BasicLevel.INFO, "Roll back a transaction: ");
        try {
            tpm.prepare(_this, false);
        } catch (PersistenceException e) {
        }
        rollbackPrepared();
    }

    public boolean isActive() {
        switch(status){
        case TransactionalPersistenceManager.CTX_ACTIVE_TRANSACTIONAL:
        case TransactionalPersistenceManager.CTX_PREPARED:
        case TransactionalPersistenceManager.CTX_PREPARED_OK:
        case TransactionalPersistenceManager.CTX_PREPARED_FAIL:
            return true;
        case TransactionalPersistenceManager.CTX_ACTIVE:
        case TransactionalPersistenceManager.CTX_COMMITTED:
        case TransactionalPersistenceManager.CTX_ABORTED:
        case TransactionalPersistenceManager.CTX_CLOSED:
        default:
            return false;
        }
    }

    public void setNontransactionalRead(boolean b) {
        nontransactionalRead = b;
    }

    public boolean getNontransactionalRead() {
        return nontransactionalRead;
    }

    public void setNontransactionalWrite(boolean b) {
        nontransactionalWrite = b;
    }

    public boolean getNontransactionalWrite() {
        return nontransactionalWrite;
    }

    public void setRetainValues(boolean b) {
        retainValues = b;
    }

    public boolean getRetainValues() {
        return retainValues;
    }

    public void setRestoreValues(boolean b) {
        restoreValues = b;
    }

    public boolean getRestoreValues() {
        return restoreValues;
    }

    public void setOptimistic(boolean b) {
        optimistic = b;
    }

    public boolean getOptimistic() {
        return optimistic;
    }

    public void setSynchronization(Synchronization s) {
        synchronization = s;
    }

    public Synchronization getSynchronization() {
        return synchronization;
    }

    public PersistenceManager getPersistenceManager() {
        return pm;
    }
}
