/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.genclass;

import org.objectweb.speedo.runtime.concurrence.optimistic.BasicSpeedoAccessor;
import org.objectweb.speedo.api.SpeedoGenClassCoherence;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.Debug;
import org.objectweb.speedo.api.SpeedoAccessor;
import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.jorm.api.PGenClassAccessor;
import org.objectweb.jorm.api.PIndexedElem;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class represents an accessor of a genclass structure.
 * This accessor is used in case of the generic class is persistent.
 *
 * @author S.Chassande-Barrioz
 */
public abstract class GenClassAccessor
        extends BasicSpeedoAccessor
        implements PGenClassAccessor, SpeedoGenClassCoherence {

    /**
     * The SpeedoProxy attached to this state representation.
     */
    public SpeedoGenClassProxy jdoProxy;

    /**
     * Indicates if this accessor support the dela mode
     */
    public boolean supportDelta = true;

    /**
     * A temporary variable which is used to search an element into the
     * "elements" list.
     */
    protected GenClassElement tmpelem = null;

	protected Logger logger;

    /**
     * The ArrayList used to store the indexed elements of the genclass.
     * Then this list contains also the deleted elements.
     */
    public List elements = null;

    public Object inner = null;

    public GenClassAccessor(SpeedoGenClassProxy jdoProxy) {
        elements = new ArrayList();
        this.jdoProxy = jdoProxy;
        this.proxy = jdoProxy;
        this.tmpelem = (GenClassElement) createPIndexedElem();
    }

    public abstract PIndexedElem createPIndexedElem(GenClassAccessor gca);

	public abstract void setGCValue(Object o);


    public Object getInner() {
        return inner;
    }

	public Logger getLogger() {
		if (logger == null) {
			logger = ((Loggable) jdoProxy).getLogger();
		}
		return logger;
	}

    public void setInner(Object inner) {
		if (Debug.ON) {
			((Loggable) jdoProxy).getLogger().log(BasicLevel.DEBUG, "setInner(" + inner + ")");
		}
		jdoProxy.jdoWriteIntention(this);
        this.inner = inner;
        if (inner != null) {
            supportDelta = false;
        }
    }

    /**
	 * To prepare the write operation the inner genclass are copied into the
	 * accessor.
	 */
	public void prepareWrite() {
		if (Debug.ON) {
			((Loggable) jdoProxy).getLogger().log(BasicLevel.DEBUG, "prepareWrite(): inner=" + inner);
		}
		if (inner != null) {
			setGCValue(inner);
			inner = null;
		}
    }

    /**
	 * At the end of the working set, the writing mode is initialized
	 * (deltaSupported), and the reference to the persistent are unswilled
	 */
	public void workingSetClosed() {
		if (Debug.ON) {
			((Loggable) jdoProxy).getLogger().log(BasicLevel.DEBUG, "workingSetClosed()");
		}
        supportDelta = true;
		Iterator it = elements.iterator();
		while (it.hasNext()) {
			GenClassElement gce = (GenClassElement) it.next();
			gce.unSwizzle();
		}
		tx = null;
    }

	public abstract void jdoDeletePersistent(ProxyManager pm);

	public abstract void jdoLoadFieldsFromAccessor(SpeedoAccessor sa);

    // IMPLEMENTATION OF THE SpeedoGenClassCoherence INTERFACE //
    //---------------------------------------------------------//

    public boolean speedoAdd(Object elemToAdd) {
        return false;
    }

    public boolean speedoRemove(Object elemToRemove) {
        return false;
    }

    // IMPLEMENTATION OF THE PGenClassAccessor INTERFACE //
    //---------------------------------------------------//

    public PIndexedElem createPIndexedElem() {
        return createPIndexedElem(this);
    }

    public Object getMemoryInstance() {
        return jdoProxy;
    }

    public void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO {
        // the elem is read from the DS, set it to unmodified.
        ((GenClassElement) elem).setStatus(PIndexedElem.ELEM_UNMODIFIED);
        elements.add(elem);
    }

    public int paGetNbElem() {
        return elements.size();
    }

    public Iterator paIterator() {
        return elements.iterator();
    }

    public boolean paDeltaSupported() {
        return supportDelta;
    }

    public void paSetNbElem(int nbelem) {
        elements = (nbelem <0 ? new ArrayList() : new ArrayList(nbelem));
    }

}
