/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.mapper;

import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.mapper.rdb.lib.MapperJDBC;
import org.objectweb.jorm.mapper.rdb.lib.ConnectionSpecJDBC;
import org.objectweb.jorm.compiler.lib.JormCompiler;
import org.objectweb.jorm.compiler.lib.JormCompilerParameter;
import org.objectweb.jorm.lib.JormConfigurator;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.UserBindingController;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.wrapper.p6spy.P6SpyLogger;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.medor.eval.prefetch.lib.PrefetchCacheImpl;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolResource;
import org.objectweb.perseus.pool.api.Pool;

import java.sql.DriverManager;
import java.sql.Driver;
import java.util.Map;
import java.util.HashMap;

public class JDBCMapper
		extends MapperJDBC
		implements UserBindingController, LifeCycleController,
		JDBCMapperAttributes, PoolMatchFactory {

    public final static String LOGGER_FACTORY_BINDING = "logger-factory";
	public final static String POOL_BINDING = "pool";
    public final static String LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".rt.mapper.jdbc";
    public final static String SPY_LOGGER_NAME
            = SpeedoProperties.LOGGER_NAME + ".rt.mapper.jdbc.sql";


	/**
	 * The pool of ConnectionResource
	 */
	protected Pool connectionPool;

	/**
	 * The association between a connection and a ConnectionResource
	 */
	protected Map connection2resource;

    private String url;

    private String userName;

    private String password;

    private String driverCN = null;

    private boolean started = false;

    private boolean initialized = false;

	private boolean poolConnection = false;

	/**
	 * destroyes a connection
	 */
	protected void destroyConnection(Object conn) throws PException {
		connection2resource.remove(conn);
		super.closeConnection(conn);
	}

	/**
	 * The class used to encapsulte the real connection in order to be used by
	 * a pool.
	 */
	public class ConnectionResource implements PoolResource {

		public Object connection;

		public boolean equals(Object obj) {
			return (obj instanceof ConnectionResource)
					&& ((ConnectionResource) obj).connection == connection;
		}

		/**
		 * <b>destroy</b> is called when this resource is removed from a Pool.
		 */
		public void destroy() throws Exception {
			destroyConnection(connection);
		}
	}


    public JDBCMapper() throws PException {
		connection2resource = new HashMap();
    }


	// IMPLEMENTATION OF THE PoolMatchFactory INTERFACE //
	//--------------------------------------------------//

	/**
	 * <b>createResource</b> creates a new PoolResource.
	 * @param hints	The "properties" that the created PoolResource should
	 *				conform to.
	 * @return		The created PoolResource.
	 */
	public PoolResource createResource(Object hints) throws Exception {
		ConnectionResource cr = new ConnectionResource();
		if (hints == null) {
			cr.connection = super.getConnection();
		} else {
			cr.connection = super.getConnection(hints);
		}
		connection2resource.put(cr.connection, cr);
		return cr;
	}

	/**
	 * <b>matchResource</b> tests if a given resource of a Pool matches with
	 * the hints passed with the Pool getResource method.
	 * @param pr	The PoolResource to test its matching with some
	 *				"properties" specified by hints.
	 * @param hints	The "properties" that the PoolResource specified by pr
	 *				should match.
	 * @return		<b>true</b> if the pr PoolResource matches the hints
	 *				"properties".
	 */
	public boolean matchResource(PoolResource pr, Object hints) {
		return true;
	}


    // IMPLEMENTATION OF THE JDBCMapperAttributes INTERFACE //
    //------------------------------------------------------//

    public String getDriverClassName() {
        return driverCN;
    }

    public void setDriverClassName(String dcn) {
        driverCN = dcn;
    }

    public String getURL() {
        return url;
    }

    public void setURL(String url) {
        this.url = url;
    }

    public String getUserName() {
        return userName;
    }

    public void setUserName(String userName) {
        this.userName = userName;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

	public void setPoolConnection(boolean pc) {
		poolConnection = pc;
	}

	public boolean getPoolConnection() {
		return poolConnection;
	}


    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String s) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            return getLoggerFactory();
		} else if (POOL_BINDING.equals(s)) {
			return connectionPool;
		} else {
        	return null;
		}
    }

    public void addFcBinding(String s, Object o) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            setLoggerFactory((LoggerFactory) o);
		} else if (POOL_BINDING.equals(s)) {
			connectionPool = (Pool) o;
		}
    }

    public void removeFcBinding(String s, Object o) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            setLoggerFactory(null);
		} else if (POOL_BINDING.equals(s)) {
			connectionPool = null;
		}
    }

    // IMPLEMENTATION OF THE LifeCycleController INTERFACE //
    //-----------------------------------------------------//

    public String getFcState() {
        return started ? STARTED : STOPPED;
    }

    public void startFc() {
        started = true;
        if (!initialized) {
            P6SpyLogger.logger = getLoggerFactory().getLogger(SPY_LOGGER_NAME);
            P6SpyLogger.level = BasicLevel.DEBUG;
            Object cf = getConnectionFactory();
            if (cf == null) {
                try {
                    DriverManager.registerDriver((Driver)
                                Class.forName(driverCN).newInstance());
                    setConnectionFactory(
                            new ConnectionSpecJDBC(url, driverCN, userName, password));
                } catch (Exception e) {
					getLogger().log(BasicLevel.ERROR,
							"Impossible to configure the jdbc access: ", e);
                    throw new RuntimeException(
                            "Impossible to configure the jdbc access: "
                            + ExceptionHelper.getNested(e).getMessage());
                }
            }
            Object o = null;
            try {
                getLogger().log(BasicLevel.DEBUG, "try to fetch a connection");
                o = getConnection();
            } catch (PException e) {
                Exception ie = ExceptionHelper.getNested(e);
                getLogger().log(BasicLevel.ERROR,
                        "Impossible to fetch a connection", ie);
				System.exit(-1);
                throw new RuntimeException(
                        "Impossible to fetch a connection: "
                        + ie.getMessage());
            } finally {
                if (o != null)  {
                    try {
                        closeConnection(o);
                    } catch (PException e) {
                    }
                }
            }

            try {
                JormCompiler jormcompiler = new JormCompiler();
                JormConfigurator jcc = jormcompiler.getCompilerConfigurator();
                jcc.configure();
                jcc.setLoggerFactory(getLoggerFactory());
                PrefetchCacheImpl pc = new PrefetchCacheImpl(
						getLoggerFactory().getLogger(
								"org.objectweb.speedo.rt.query.prefetch"));
				setPrefetchCache(pc);
                setJormConfigurator(jcc);
                start();
                initialized = true;
            } catch (PException e) {
                throw new RuntimeException(
                        "Impossible to configure the mapper: "
                        + ExceptionHelper.getNested(e).getMessage());
            }
        }
    }

    public void stopFc() {
        started = false;
    }


	// IMPLEMENTATION OF THE PMapper INTERFACE //
	//-----------------------------------------//

	/**
	 * The pool is used to fetch a connection.
	 */
	public Object getConnection() throws PException {
		if (poolConnection) {
			try {
				ConnectionResource cr = (ConnectionResource)
						connectionPool.getResource(null);
				return cr.connection;
			} catch (Exception e) {
				throw new PExceptionIO(ExceptionHelper.getNested(e),
						"Impossible to fetch a jdbc connection on driver");
			}
		} else {
			return super.getConnection();
		}
	}

	/**
	 * The connection is release into the pool
	 * @param conn
	 * @throws PException
	 */
	public void closeConnection(Object conn) throws PException {
		if (poolConnection) {
			if (conn == null) {
				return;
			}
			try {
				ConnectionResource cr = (ConnectionResource)
						connection2resource.get(conn);
				if (cr != null) {
					connectionPool.releaseResource(cr);
				}
			} catch (Exception e) {
				throw new PExceptionIO(ExceptionHelper.getNested(e),
						"Impossible to release a jdbc connection");
			}
		} else {
			super.closeConnection(conn);
		}
	}
}
