/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: X. Spengler
 *
 */

package org.objectweb.speedo.runtime.query.parser;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.Bitwize;
import org.objectweb.medor.filter.lib.ConditionalAnd;
import org.objectweb.medor.filter.lib.ConditionalOr;
import org.objectweb.medor.filter.lib.DivideBy;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.Greater;
import org.objectweb.medor.filter.lib.GreaterEqual;
import org.objectweb.medor.filter.lib.Lower;
import org.objectweb.medor.filter.lib.LowerEqual;
import org.objectweb.medor.filter.lib.Minus;
import org.objectweb.medor.filter.lib.Mult;
import org.objectweb.medor.filter.lib.Not;
import org.objectweb.medor.filter.lib.NotEqual;
import org.objectweb.medor.filter.lib.Or;
import org.objectweb.medor.filter.lib.Plus;
import org.objectweb.medor.filter.lib.UMinus;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Stack;
import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of a visitor that creates the filter
 */
public class SpeedoQLQueryFilterVisitor extends SpeedoQLAbstractVisitor {

    private final static String CONTAINS_PATH_SET = "Next element is the contain path set";
    private final static String REMOVER = "Remover";


    /**
     * qf is the transformation of the filter into a MEDOR expression
     */
    private Expression qf = null;

    /**
     * qt is built by the SpeedoQLVariableVisitor and is just used here
     */
    private QueryTree qt = null;

    /**
     * speedoql is the result of the filter parsing
     */
    private ASTSpeedoQL speedoql = null;

    private String tab ="";

    private boolean debug = false;

    /**
     * Constructor
     * @param qtree a QueryTree built with the list of variables, parameters and name.
     * @param speedoql root of the lexical tree of the query
     */
    public SpeedoQLQueryFilterVisitor(QueryTree qtree,
                                      ASTSpeedoQL speedoql) {
        qt = qtree;
        this.speedoql = speedoql;
    }

    /**
     * the visit starts here, please before sets a list of variables.
     * @throws Exception
     */
    public void startVisiting() throws Exception {
        debug = logger.isLoggable(BasicLevel.DEBUG);
        try {
            visit(speedoql);
        } catch (VisitorException e) {
            throw e.getNestedException();
        }
    }

    /**
     * get the query filter that was built from visiting the syntaxic tree
     */
    public Expression getQueryFilter() {
        return qf;
    }

    public Object visit(ASTPrimary node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTSpeedoPrimary node, Object data) {
        visit((SimpleNode) node, data);
        Stack stack = (Stack) data;
        while(stack.size() >0 && REMOVER.equals(stack.peek())) {
            stack.pop();
        }
        qf = (Expression) stack.pop();
        return null;
    }

    /**
     *
     */
    public Object visit(ASTRelationalExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit RelationalExpression: " + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0) {
            if (CONTAINS_PATH_SET.equals(stack.peek())) {
                if (debug) {
                    logger.log(BasicLevel.ERROR, "remove the element of the stack");
                }
            } else {
                if (debug) {
                    logger.log(BasicLevel.DEBUG, "manage relational expression: "
                        + "(children: " + node.jjtGetNumChildren()
                        + ", stack: " + stack.size()
                        + ", peek:" + stack.peek() + ")");
                }
                for (int i = 0; (i < (node.jjtGetNumChildren() - 1) && stack.size() > 0); i++) {
                    int op = ((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue();
                    if (usedInRelationalExpresssion(op)) {
                        Object child2 = stack.pop();
                        Object child1 = stack.pop();

                        if (debug) {
                            logger.log(BasicLevel.DEBUG, "pop child1: " + child1);
                            logger.log(BasicLevel.DEBUG, "pop child2: " + child2);
                        }
                        if (REMOVER.equals(child1)) {
                            if (REMOVER.equals(child2)) {
                                //nothin
                            } else {
                                stack.push(child2);
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + child2 + ")");
                                }
                            }
                        } else {
                            if (REMOVER.equals(child2)) {
                                stack.push(child1);
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + child1 + ")");
                                }
                            } else {
                                Expression ret = null;
                                switch (op) {
                                case SpeedoQLConstants.OR:
                                    ret = new ConditionalOr((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.AND:
                                    ret = new ConditionalAnd((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.BITWISEOR:
                                    ret = new Or((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.BITWISEXOR:
                                    ret = null;
                                    break;
                                case SpeedoQLConstants.BITWISEAND:
                                    ret = new And((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.EQ:
                                    ret = new Equal((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.NEQ:
                                    ret = new NotEqual((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.LT:
                                    ret = new Lower((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.GT:
                                    ret = new Greater((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.GTE:
                                    ret = new GreaterEqual((Expression) child1, (Expression) child2);
                                    break;
                                case SpeedoQLConstants.LTE:
                                    ret = new LowerEqual((Expression) child1, (Expression) child2);
                                    break;
                                }
                                if (debug) {
                                    logger.log(BasicLevel.DEBUG, "push(" + ret + ")");
                                }
                                stack.push(ret);
                            }
                        }
                    }
                }
            }
        }
        logger.log(BasicLevel.DEBUG, "children:" + node.jjtGetNumChildren()
                + " / stack: " + stack.size());
        logger.log(BasicLevel.DEBUG, tab + "End of Visit RelationalExpression: " + node);
        return null;
    }

    private boolean usedInRelationalExpresssion(int op) {
        switch (op) {
        case SpeedoQLConstants.OR:
        case SpeedoQLConstants.AND:
        case SpeedoQLConstants.BITWISEOR:
        case SpeedoQLConstants.BITWISEXOR:
        case SpeedoQLConstants.BITWISEAND:
        case SpeedoQLConstants.EQ:
        case SpeedoQLConstants.NEQ:
        case SpeedoQLConstants.LT:
        case SpeedoQLConstants.GT:
        case SpeedoQLConstants.GTE:
        case SpeedoQLConstants.LTE:
            logger.log(BasicLevel.DEBUG, "node useful");
            return true;
        default:
            logger.log(BasicLevel.DEBUG, "node useless");
            return false;
        }
    }

    public Object visit(ASTAdditiveExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit AdditiveExpression: " + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0
                && !CONTAINS_PATH_SET.equals(stack.peek())
                && !REMOVER.equals(stack.peek())) {
            Expression ret = (Expression) stack.pop();
            for (int i = 0; i < node.jjtGetNumChildren() - 1; i++) {
                logger.log(BasicLevel.DEBUG, "Visit ConditionalExpression... children...[" + i + "]");

                switch (((Integer) node.ops.get(node.jjtGetNumChildren() - 2 - i)).intValue()) {

                case SpeedoQLConstants.PLUS:
                    ret = new Plus((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.MINUS:
                    ret = new Minus((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.MULT:
                    ret = new Mult((Expression) stack.pop(), ret);
                    break;
                case SpeedoQLConstants.DIV:
                    ret = new DivideBy((Expression) stack.pop(), ret);
                    break;
                default:

                }
            }
            ((Stack) data).push(ret);
        }
        logger.log(BasicLevel.DEBUG, tab + "End of Visit AdditiveExpression: " + node);
        return null;
    }

    public Object visit(ASTUnaryExpression node, Object data) {
        logger.log(BasicLevel.DEBUG, tab + "Visit UnaryExpression" + node);
        tab += '\t';
        visit((SimpleNode) node, data);
        tab = tab.substring(1);
        Stack stack = (Stack) data;
        if (stack.size() > 0) {
            logger.log(BasicLevel.DEBUG, "stack.peek:" + stack.peek());
        }
        if (stack.size() > 0
                && !CONTAINS_PATH_SET.equals(stack.peek())
                && !REMOVER.equals(stack.peek())) {
            Object o = stack.pop();
            if (node.ops.size() > 0) {
                switch (((Integer) node.ops.get(0)).intValue()) {
                case SpeedoQLConstants.MINUS:
                    o = new UMinus((Expression) o);
                    break;
                case SpeedoQLConstants.BITWISECOMPL:
                    o = new Bitwize((Expression) o);
                    break;
                case SpeedoQLConstants.NOT:
                    logger.log(BasicLevel.DEBUG, "NOT(" + o + "): " + node);
                    o = new Not((Expression) o);
                    break;
                }
            }
            ((Stack) data).push(o);
        }
        logger.log(BasicLevel.DEBUG, tab + "End of Visit UnaryExpression: " + node);
        return null;
    }

    // to be done
    public Object visit(ASTCastExpression node, Object data) {
        ((Stack) data).push(node);
        return null;
    }

    /**
     * 4 cases to manage: (en cours par equipe MEDOR)
     * - Collection.contains(Object o)
     * - Collection.isEmpty()
     * - String.startsWith(String s)
     * - String.endsWith(String s)
     */
    public Object visit(ASTArgumentList node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTLiteral node, Object data) {
        if (node.value instanceof Integer) {
            ((Stack) data).push(new BasicOperand(((Integer) node.value).intValue()));
        } else if (node.value instanceof Float) {
            ((Stack) data).push(new BasicOperand(((Float) node.value).floatValue()));
        } else if (node.value instanceof Character) {
            ((Stack) data).push(new BasicOperand(((Character) node.value).charValue()));
        } else if (node.value instanceof String) {
            String s = (String) (node.value);
            s = s.substring(1, s.length()-1);
            ((Stack) data).push(new BasicOperand(s));
        } else if (node.value instanceof Boolean) {
            ((Stack) data).push(new BasicOperand(((Boolean) node.value).booleanValue()));
        }
        // what do I do with a null value ??????????
        //	else ((Stack) data).push(new BasicOperand(null));

        return null;
    }

    public Object visit(ASTType node, Object data) {
        logger.log(BasicLevel.DEBUG, "Visit Type: " + node);

        ((Stack) data).push(node);

        logger.log(BasicLevel.DEBUG, "End of Visit Type");
        return null;
    }

    /**
     * qualifiedname could be:
     * - a class field (salary from Employee)
     * - a parameter
     * - a variable
     *
     */
    public Object visit(ASTQualifiedName node, Object data) {
        String name = (String) node.value;
        String[] splitted = splitPath(name);
        Stack stack = (Stack) data;
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Visit QualifiedName value=["
                    + node.value + "] stack size=" + stack.size()
                    + " /splitted: [" + buildStringwithout(splitted, -1, ".") + "]");
        }
        try {
            if (params != null && params.get(splitted[0]) != null) {
                if (stack.size() > 0 && CONTAINS_PATH_SET.equals(stack.peek())) {
                    logger.log(BasicLevel.DEBUG, "pop: " + stack.pop());
                    String pathset = (String) stack.pop() + "."
                            + NavigatorNodeFactory.GENCLASS_ELEMENT_NAME;
                    logger.log(BasicLevel.DEBUG,
                            "create a member of between gen class element '"
                            + pathset + "' and the parameter '" + name + "'");
                    QueryTreeField qtf =  (QueryTreeField)
                            qt.getTupleStructure().getField(pathset);
                    Expression e = new MemberOf(
                            singletonList(params.get(name)),
                            singletonList(new BasicFieldOperand(qtf)));
                    logger.log(BasicLevel.DEBUG, "push(" + ExpressionPrinter.e2str(e) + ")");
                    stack.push(e);
                } else {
                    logger.log(BasicLevel.DEBUG, "create a parameterOperand with:" + name);
                    stack.push(params.get(name));
                }
            } else if (vars != null && vars.get(splitted[0]) != null) {
                if (stack.size() > 0 && CONTAINS_PATH_SET.equals(stack.peek())) {
                    logger.log(BasicLevel.DEBUG, "Do not use the pathset of the container operator");
                    stack.pop(); // pop the CONTAINS_PATH_SET
                    stack.pop(); // pop the pathset
                    stack.push(REMOVER);
                } else {
                    logger.log(BasicLevel.DEBUG, "create a variableOperand with:" + name);
                    Field f = qt.getTupleStructure().getField(name);
                    stack.push(new BasicFieldOperand(f));
                }

            } else if (splitted.length >= 2
                    && "isEmpty".equals(splitted[splitted.length - 1])) {
                logger.log(BasicLevel.DEBUG, "Visit IsEmpty: " + node);
                stack.push(REMOVER);

            } else if ("startsWith".equals(splitted[splitted.length - 1])) {
//TODO: Support the startsWith operator
                logger.log(BasicLevel.ERROR, "The startsWith operator is not yet implemented");

            } else if ("endsWith".equals(splitted[splitted.length - 1])) {
//TODO: Support the endsWith operator
                logger.log(BasicLevel.ERROR, "The endsWith operator is not yet implemented");

            } else if (splitted.length >= 2
                    && "contains".equals(splitted[splitted.length - 1])) {
                String pathset = curClass + "." + withoutThis(
                        buildStringwithout(splitted, splitted.length-1, "."));
                logger.log(BasicLevel.DEBUG, "contains operator: set=" + pathset);
                stack.push(pathset);
                stack.push(CONTAINS_PATH_SET);
            } else {
                String theName = curClass + "." + withoutThis(name);
                logger.log(BasicLevel.DEBUG, "create a fieldOperand with:" + theName);
                Field f = qt.getTupleStructure().getField(theName);
                stack.push(new BasicFieldOperand(f));
            }
        } catch (Exception e) {
            throw new VisitorException(e);
        }
        return null;
    }
    private static List singletonList(Object element) {
        List res = new ArrayList(1);
        res.add(element);
        return res;
    }
}
